#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>

#include "defines.h"
#include "DOTReaderWriter.h"

using namespace std;

inline int round(float f)
{
    return (int)(f + (f > 0.0 ? 0.5 : -0.5));
}

DOTReaderWriter::DOTReaderWriter(char* in_file, char* out_file)
: m_in_file(in_file),
  m_out_file(out_file),
  m_num_nodes(0)
{
}

unsigned int DOTReaderWriter::read(unsigned char** &graph)
{
    string line;
    ifstream fs;

    fs.open(m_in_file);
    if (fs.is_open())
    {
        printf("Using graph in file %s\n", m_in_file);
        if (DEBUG) printf("----------\n");

        // detect graph type
        while (!fs.eof())
        {
            getline(fs, line);
            if (line.find("digraph") != string::npos)
            {
                m_edge_str = "->";
                break;
            }
            else if (line.find("graph") != string::npos)
            {
                m_edge_str = "--";
                break;
            }
        }

        // get list of nodes
        while (!fs.eof())
        {
            getline(fs, line);

            bool done = false;
            size_t i = 0;
            size_t j = line.find(m_edge_str);
            while (j != string::npos)
            {
                string node_str = line.substr(i, j - i);

                // trim
                size_t a = node_str.find_first_not_of(" \t\"");
                size_t b = node_str.find_last_not_of(" \t\"") + 1;
                node_str = node_str.substr(a, b - a);

                if (m_node_map.find(node_str) == m_node_map.end())
                {
                    if (DEBUG) printf("%d: %s\n", m_num_nodes, node_str.c_str());
                    m_node_vec.push_back(node_str);
                    m_node_map[node_str] = m_num_nodes++;
                }

                if (done) break;

                i = j + 2;
                j = line.find(m_edge_str, i);
                if (j == string::npos)
                {
                    // last node of the line
                    done = true;

                    j = line.find_first_of("[;", i);
                    if (j == string::npos) j = line.length();
                }
            }
        }

        // populate graph
        graph = (unsigned char**)malloc(m_num_nodes * sizeof(unsigned char*));
        for (unsigned int i = 0; i < m_num_nodes; ++i)
        {
            graph[i] = (unsigned char*)malloc(m_num_nodes * sizeof(unsigned char));
            memset(graph[i], (unsigned char)0, m_num_nodes * sizeof(unsigned char));
        }
        fs.close();
        fs.open(m_in_file);
        while (!fs.eof()) // advance pointer to graph definition
        {
            getline(fs, line);
            if (line.find("digraph") != string::npos) break;
            else if (line.find("graph") != string::npos) break;
        }
        while (!fs.eof())
        {
            getline(fs, line);

            size_t i = 0;
            size_t j = line.find(m_edge_str);
            if (j != string::npos)
            {
                bool done = false;
                string node_str = line.substr(i, j - i);
                size_t a = node_str.find_first_not_of(" \t\"");
                size_t b = node_str.find_last_not_of(" \t\"") + 1;
                node_str = node_str.substr(a, b - a);

                unsigned int last_node_index = m_node_map[node_str];
                while (!done)
                {
                    i = j + 2;
                    j = line.find(m_edge_str, i);
                    if (j == string::npos)
                    {
                        // last node of the line
                        done = true;

                        j = line.find_first_of("[;", i);
                        if (j == string::npos) j = line.length();
                    }

                    node_str = line.substr(i, j - i);
                    size_t a = node_str.find_first_not_of(" \t\"");
                    size_t b = node_str.find_last_not_of(" \t\"") + 1;
                    node_str = node_str.substr(a, b - a);
                    int cur_node_index = m_node_map[node_str];

                    // update graph
                    graph[last_node_index][cur_node_index]++;
                    graph[cur_node_index][last_node_index]++;
                    if (DEBUG) printf("%d %s %d\n", last_node_index, m_edge_str.c_str(), cur_node_index);

                    last_node_index = cur_node_index;
                }
            }
        }

        fs.close();
    }

    return m_num_nodes;
}

void DOTReaderWriter::write(float* pos_x, float* pos_y)
{
    string line;
    ifstream ifs;
    ofstream ofs;

    ifs.open(m_in_file);
    ofs.open(m_out_file);
    if (!ifs.is_open())
    {
        printf("Error opening file %s\n", m_in_file);
        exit(1);
    }
    if (!ofs.is_open())
    {
        printf("Error opening file %s\n", m_out_file);
        exit(1);
    }

    // detect nodes with node statement
    bool* node_stmt_exists = (bool*)malloc(m_num_nodes * sizeof(bool));
    memset(node_stmt_exists, false, m_num_nodes * sizeof(bool));

    while (!ifs.eof()) // advance pointer to graph definition
    {
        getline(ifs, line);
        ofs << line << endl;
        if (line.find("{") != string::npos) break;
    }
    while (!ifs.eof())
    {
        getline(ifs, line);

        if (line.find(m_edge_str) == string::npos)
        {
            // not edge statement
            if (line.find("}") != string::npos) break;

            // this is a node statement
            size_t i = line.find("[");
            if (i == string::npos)
            {
                if (line.find_first_of(" \t") != string::npos)
                    printf("Error: I don't understand this line \"%s\"\n", line.c_str());
            }
            else
            {
                string node_str = line.substr(0, i);

                // trim
                size_t a = node_str.find_first_not_of(" \t\"");
                size_t b = node_str.find_last_not_of(" \t\"") + 1;
                node_str = node_str.substr(a, b - a);

                if (m_node_map.find(node_str) == m_node_map.end())
                {
                    printf("Found disconnected node \"%s\"\n", node_str.c_str());
                }
                else
                {
                    unsigned int node_index = m_node_map[node_str];
                    node_stmt_exists[node_index] = true;
                    char pos_str[1024];
                    sprintf(pos_str, "pos=\"%d,%d\",", round(pos_x[node_index]), round(pos_y[node_index]));
/*
                    string pos_str = "pos=\"";
		    pos_str += (int)pos_x[node_index];
		    pos_str += ",";
		    pos_str += (int)pos_y[node_index];
		    pos_str += "\",";
*/
                    line.insert(i + 1, pos_str);
                }
            }
        }

        ofs << line << endl;
	line.clear();
    }

    // add the remaining nodes
    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        if (!node_stmt_exists[i])
            ofs << "\t\"" << m_node_vec[i] << "\" [pos=\"" << round(pos_x[i]) << "," << round(pos_y[i]) << "\",label=\"" << i << "\",height=0.5,width=0.5,fixedsize=true];" << endl;
    }

    ofs << "\tgraph [splines=true];" << endl;

    if (!line.empty())
        ofs << line << endl;

    while (!ifs.eof())
    {
        getline(ifs, line);
	ofs << line << endl;
    }

    ifs.close();
    ofs.close();
}
