#include "FDPSimulator.h"

#include <assert.h>
#include <memory.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <cutil.h>

//Since this was written in C++ need to expose all the calls into C
extern "C" void basic_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void float2_fps_gpu_algorithm(unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void proper_reduction_fps_gpu_algorithm(unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void force_calc_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void force_calc_kin_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void local_memory_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void data_size_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void matrix_transpose_fps_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void use_grid_blocks_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void use_grid_blocks_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void force_reduction_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);
extern "C" void float4_forces_gpu_algorithm( unsigned int num_nodes, unsigned char** graph, float* x_pos, float* y_pos);

static const char* ALG_NAMES[] = {
	"CPU Force-based implementation",
	"Basic GPU  implementation",
	"GPU implementation using float2 rather then straight floats",
	//"Use Proper Reduction Technique",
	"GPU: Parallelize Force Based Calc",
	"GPU: Parallelize Force Based Calc and Kinetic Energy",
	"Use local memory",
	"Use int instead of char",
	"Matrix Transpose",
	"Use Grid Blocks",
	"Force reduction kernel",
	"Float 4 forces"
 };

FDPSimulator::FDPSimulator(unsigned char** graph, unsigned int num_nodes)
: m_num_nodes(num_nodes),
  m_graph(graph),
  m_damping(DAMPING),
  m_timestep(TIMESTEP)
{
    m_pos_x = (float*)malloc(num_nodes * sizeof(float));
    m_pos_y = (float*)malloc(num_nodes * sizeof(float));
    m_vel_x = (float*)malloc(num_nodes * sizeof(float));
    m_vel_y = (float*)malloc(num_nodes * sizeof(float));
    m_force_x = (float*)malloc(num_nodes * sizeof(float));
    m_force_y = (float*)malloc(num_nodes * sizeof(float));
}

FDPSimulator::~FDPSimulator()
{
    free(m_pos_x);
    free(m_pos_y);
    free(m_vel_x);
    free(m_vel_y);
    free(m_force_x);
    free(m_force_y);
}

void FDPSimulator::init()
{
//   srand48((long)time(NULL));
   srand48(0);
   for (unsigned int i = 0; i < m_num_nodes; ++i)
   {
       m_pos_x[i] = (float)(drand48() * m_num_nodes);
       m_pos_y[i] = (float)(drand48() * m_num_nodes);
   }
   
    memset(m_vel_x, 0, m_num_nodes * sizeof(float));
    memset(m_vel_y, 0, m_num_nodes * sizeof(float));   
}

const char* FDPSimulator::get_name
(
	FDPSimulator::ALGORITHM alg
) const
{
	assert ( (unsigned int )alg < sizeof(ALG_NAMES)/sizeof(ALG_NAMES[0]) );
	return ALG_NAMES[alg];
}

void FDPSimulator::update_force()
{
    float force_x = 0, force_y = 0;

    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        m_force_x[i] = 0;
        m_force_y[i] = 0;

        for (unsigned int j = 0; j < m_num_nodes; ++j)
        {
            if (i != j)
            {
                calc_force(m_pos_x[i], m_pos_y[i], m_pos_x[j], m_pos_y[j], m_graph[i][j], &force_x, &force_y);
                m_force_x[i] += force_x;
                m_force_y[i] += force_y;
            }
        }
    }
}

void FDPSimulator::update_vel()
{
    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        m_vel_x[i] += m_timestep * m_force_x[i];
        m_vel_y[i] += m_timestep * m_force_y[i];
        m_vel_x[i] *= m_damping;
        m_vel_y[i] *= m_damping;
    }
}

void FDPSimulator::update_pos()
{
    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        m_pos_x[i] += m_timestep * m_vel_x[i];
        m_pos_y[i] += m_timestep * m_vel_y[i];
    }
}

float FDPSimulator::get_kinetic_energy()
{
    float result = 0;

    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        result += m_vel_x[i] * m_vel_x[i] + m_vel_y[i] * m_vel_y[i];
    }

    return result;
}

void FDPSimulator::simulate
(
	ALGORITHM alg 
)
{
    if (alg == CPU)
    {
	    unsigned int num_iter = 0;
            float kinetic_energy = 0;
	    do
	    {
	        update_force();
	        update_vel();
        	update_pos();
                kinetic_energy = get_kinetic_energy();
                if (DEBUG) if (num_iter % 10000 == 1) printf("Energy = %f\n", kinetic_energy);
	        ++num_iter;
	    }
	    while (kinetic_energy > STABLE_ENERGY_THRESHOLD * m_num_nodes);
	    printf("Total iterations: %d\n", num_iter);
            printf("Energy = %f\n", kinetic_energy);
    } 
    else if (alg == GPU_BASIC)
    {
    	basic_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == GPU_USE_FLOAT2)
    {
    	float2_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == PROPER_REDUCTION)
    {
    	proper_reduction_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == GPU_FORCE_CALC)
    {
        force_calc_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == GPU_FORCE_CALC_KIN)
    {
        force_calc_kin_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == LOCAL_MEMORY)
    {
        local_memory_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == DATA_SIZE)
    {
        data_size_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == MATRIX_TRANSPOSE)
    {
        matrix_transpose_fps_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if ( alg == USE_GRID_BLOCKS)
    {
	use_grid_blocks_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if ( alg == FORCE_REDUCTION)
    {
	force_reduction_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else if (alg == FLOAT4_FORCES)
    {
	float4_forces_gpu_algorithm(m_num_nodes, m_graph, m_pos_x, m_pos_y);
    }
    else {
        printf("Unknown Algorithm Type!\n");
        exit(1);
    }
}

void FDPSimulator::calc_force(float this_x, float this_y, float other_x, float other_y, unsigned char edge_weight, float* force_x, float* force_y)
{
    float dist_x = other_x - this_x;
    float dist_y = other_y - this_y;
    float dist_squared = dist_x * dist_x + dist_y * dist_y;
    float dist = sqrt(dist_squared);

    // repulsive force (charge)
    float normalized_force = REPULSIVE_FORCE_CONSTANT / dist_squared / dist;
    *force_x = -dist_x * normalized_force;
    *force_y = -dist_y * normalized_force;

    // attractive force (spring), natural length = 100
    *force_x += ATTRACTIVE_FORCE_CONSTANT * (dist_x - dist_x * 100 / dist) * edge_weight;
    *force_y += ATTRACTIVE_FORCE_CONSTANT * (dist_y - dist_y * 100 / dist) * edge_weight;
}

void FDPSimulator::dump()
{
    float max_x = m_pos_x[0];
    float min_x = m_pos_x[0];
    float max_y = m_pos_y[0];
    float min_y = m_pos_y[0];

    for (unsigned int i = 1; i < m_num_nodes; ++i)
    {
        if (m_pos_x[i] > max_x) max_x = m_pos_x[i];
        if (m_pos_x[i] < min_x) min_x = m_pos_x[i];
        if (m_pos_y[i] > max_y) max_y = m_pos_y[i];
        if (m_pos_y[i] < min_y) min_y = m_pos_y[i];
    }
#ifdef false
    unsigned int range_x = 80;
    float factor_x = (float)range_x / (max_x - min_x);
    float factor_y = factor_x * 7 / 12;
    unsigned int range_y = (unsigned int)((max_y - min_y) * factor_y) + 1;
    range_x++;


    unsigned int** graph = (unsigned int**)malloc(range_x * sizeof(unsigned int));
    for (unsigned int i = 0; i < range_x; ++i)
    {
        graph[i] = (unsigned int*)malloc(range_y * sizeof(unsigned int));
        memset(graph[i], 0, m_num_nodes * sizeof(unsigned int));
    }
    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        graph[(int)((m_pos_x[i] - min_x) * factor_x)][(int)((m_pos_y[i] - min_y) * factor_y)] = i + 1;
    }
#endif
    printf("==========\n");
    for (unsigned int i = 0; i < m_num_nodes; ++i)
    {
        //printf("%d: %d, %d\n", i, (int)((m_pos_x[i] - min_x) * factor_x), (int)((m_pos_y[i] - min_y) * factor_y));
        printf("%d: %f, %f\n", i + 1, m_pos_x[i], m_pos_y[i]);
    }
#ifdef false
    printf("----------\n");
    for (unsigned int y = 0; y < range_y; ++y)
    {
        for (unsigned int x = 0; x < range_x; ++x)
        {
            if (graph[x][y] == 0) printf(" ");
            else printf("%d", graph[x][y]);
        }
        printf("|\n");
    }
#endif
}

