#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"
#include "reduction_force.cu"

// sum up columns instead of rows
#define FORCE_XPOSE 1

// use reduction to add forces, set FORCE_XPOSE to 0 if using reduction
// WARNING! this is extremely slow for some reason...
#define FORCE_REDUCTION 0

#define SHOW_TIME_BREAKDOWN


// This kernel computes the forces on the nodes (each thread computes the forces
// between a pair of nodes)
__global__ void use_grid_blocks_gpu_kernel ( 	unsigned int* graph_data,
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					int N
				   )
{
	int x_index = (blockIdx.x * blockDim.x) + threadIdx.x;
	int y_index = (blockIdx.y * blockDim.y) + threadIdx.y;
	int index = y_index * N + x_index;
	unsigned int weight;
	int valid = (x_index < N && y_index < N);

	float2 my_r_force;
	float2 my_a_force;

	__shared__ float3 local_pos_x[16];
	__shared__ float3 local_pos_y[16];

	//first we get the 16 by 16 grid we are going to be working on use the 0th and 2nd thread
	//so they come from sepearte warps.
	if (threadIdx.y == 0 )
	{
		float2 tmp = pos[ (blockIdx.x * blockDim.x) + threadIdx.x ];
		local_pos_x[threadIdx.x].x = tmp.x;
		local_pos_x[threadIdx.x].y = tmp.y;
	}
	else if (threadIdx.y == 2)
	{
		//magic since we use the x to iterate over all the threads
		float2 tmp = pos[ (blockIdx.y * blockDim.y) + threadIdx.x ];
		local_pos_y[threadIdx.x].x = tmp.x;
		local_pos_y[threadIdx.x].y = tmp.y;
	}
	if ( valid)
	{
		weight = graph_data[index];
	}

	my_r_force.x = 0.0f;
	my_r_force.y = 0.0f;
	my_a_force.x = 0.0f;
	my_a_force.y = 0.0f;
	__syncthreads();

	if (valid)
	{

		float3 n1_pos = local_pos_x[threadIdx.x];
		float3 n2_pos = local_pos_y[threadIdx.y];

		if (x_index != y_index)
		{
			// Compute distance between Node1 and Node2
			float dist_x = n2_pos.x - n1_pos.x;
			float dist_y = n2_pos.y - n1_pos.y;
			float dist_squared = dist_x * dist_x + dist_y * dist_y;
			float dist = sqrt(dist_squared);

			// Compute Repulsive Force (but without REPULSIVE_FORCE_CONSTANT)
			my_r_force.x = -dist_x / dist_squared / dist;
			my_r_force.y = -dist_y / dist_squared / dist;

			// Compute Attractive Force (but without ATTRACTIVE_FORCE_CONSTANT)
			if (weight != 0)
			{
				my_a_force.x = (dist_x - dist_x * 100 / dist) * weight;
				my_a_force.y = (dist_y - dist_y * 100 / dist) * weight;
			}
		}

		r_force[index] = my_r_force;
		a_force[index] = my_a_force;
	}
}


// Calculates the position of each node (based on forces already pre-computed)
__global__ void use_grid_blocks_pos_gpu_kernel (
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					float2* vel,
					float2* final_pos,
					float* d_kenetic_energy_per_node,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
		float2 my_pos = pos[index];
		float2 my_vel = vel[index];

		float2 my_r_force;
		float2 my_a_force;
		my_r_force.x = 0.0f;
		my_r_force.y = 0.0f;
		my_a_force.x = 0.0f;
		my_a_force.y = 0.0f;

		// Compute force on this node
		for (int i = index; i < N*N; i+=N)
		{
			float2 r = r_force[i];
			float2 a = a_force[i];
			my_r_force.x += r.x;
			my_r_force.y += r.y;
			my_a_force.x += a.x;
			my_a_force.y += a.y;
		}

		my_r_force.x *= REPULSIVE_FORCE_CONSTANT;
		my_r_force.y *= REPULSIVE_FORCE_CONSTANT;
		my_a_force.x *= ATTRACTIVE_FORCE_CONSTANT;
		my_a_force.y *= ATTRACTIVE_FORCE_CONSTANT;

		// Computer Velocities on this node
		my_vel.x = (my_vel.x + (TIMESTEP * (my_r_force.x + my_a_force.x))) * DAMPING;
		my_vel.y = (my_vel.y + (TIMESTEP * (my_r_force.y + my_a_force.y))) * DAMPING;

		vel[index] = my_vel;

		// Compute Final Position of this node
		float2 my_final_pos;
		my_final_pos.x = my_pos.x + TIMESTEP * my_vel.x;
		my_final_pos.y = my_pos.y + TIMESTEP * my_vel.y;			
		final_pos[index] = my_final_pos;

		// Compute the Kinetic Energy of this node
		d_kenetic_energy_per_node[index] = my_vel.x * my_vel.x + my_vel.y * my_vel.y;
	}
}


extern "C" void use_grid_blocks_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float2 *d_rforce, *d_aforce, *d_pos, *d_vel, *d_final_pos;
	unsigned int* d_graph_data;
	unsigned int* mapped_graph_data;
	float2 *h_vel, *h_pos;
	float *d_kenetic_energy_per_node;
	float *h_kenetic_energy_per_node;

	unsigned int p1timer, p2timer;

	cutCreateTimer(&p1timer);
	cutCreateTimer(&p2timer);
	
	//add a bit of extra space on so we don't it ends of array reading data.
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_pos, sizeof(float2) * (num_nodes + 15) ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_vel, sizeof(float2) * (num_nodes + 15) ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_pos, sizeof(float2) * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_rforce, sizeof(float2) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_aforce, sizeof(float2) * num_nodes * num_nodes));
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned int) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy_per_node, sizeof(float) * num_nodes) );
	h_kenetic_energy_per_node = (float*)malloc(sizeof(float) * num_nodes);
		
	//translate all the data into the formats we are using in CUDA.
	mapped_graph_data = (unsigned int*)malloc(num_nodes * num_nodes * sizeof(unsigned int) );
	h_vel = (float2*)malloc(sizeof(float2) * num_nodes);
	h_pos = (float2*)malloc(sizeof(float2) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_vel[i].x = 0.0f;
		h_vel[i].y = 0.0f;
		
		h_pos[i].x = x_pos[i];
		h_pos[i].y = y_pos[i];
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned int) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_vel, (void*) h_vel, sizeof(float2) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_pos, (void*) h_pos, sizeof(float2) * num_nodes, cudaMemcpyHostToDevice));

	int counter = 0;
	float kenetic_energy = 0.0f;

	dim3 first_kernel_blocks = dim3 ( (num_nodes + 15) /  16, (num_nodes + 15) /  16 );
	if (DEBUG) printf ("Creating force grid with %d edges\n", (num_nodes + 15) /  16);
	do 
	{
#ifdef SHOW_TIME_BREAKDOWN
		cutStartTimer(p1timer);
#endif

		use_grid_blocks_gpu_kernel<<< first_kernel_blocks, dim3(16,16)>>>(d_graph_data, d_rforce, d_aforce, d_pos, num_nodes);						

		cudaThreadSynchronize();

#ifdef SHOW_TIME_BREAKDOWN
	    	cutStopTimer(p1timer);	
		cutStartTimer(p2timer);
#endif

		use_grid_blocks_pos_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_rforce, d_aforce, d_pos, d_vel, d_final_pos, d_kenetic_energy_per_node, num_nodes);

		cudaThreadSynchronize();

		// Compute the Kinetic Energy of the System
		//do_reduction(d_vel, num_nodes, &kenetic_energy);
		CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_kenetic_energy_per_node, (void*) d_kenetic_energy_per_node, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		kenetic_energy = 0.0f;
		for (unsigned int i = 0; i < num_nodes; ++i) {
			kenetic_energy += h_kenetic_energy_per_node[i];
		}
#ifdef SHOW_TIME_BREAKDOWN
	    	cutStopTimer(p2timer);
#endif
		//swap the position buffers
		float2 *tmp;
		tmp = d_pos;
		d_pos = d_final_pos;
		d_final_pos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while ( kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);

#ifdef SHOW_TIME_BREAKDOWN
    	printf("P1 timer time was  %f\n", cutGetTimerValue(p1timer));
    	printf("P2 timer time was  %f\n", cutGetTimerValue(p2timer));
#endif
	
	//restore the points to the correct buffers (d_pos contains the last result because pointers were swapped)
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_pos, (void*) d_pos, sizeof(float2) * num_nodes, cudaMemcpyDeviceToHost));
	for (unsigned int i = 0; i < num_nodes; ++i)
	{
		x_pos[i] = h_pos[i].x;
		y_pos[i] = h_pos[i].y;
	}
	
	
	free(h_kenetic_energy_per_node);
	free(mapped_graph_data);
	free(h_vel);
	free(h_pos);
	cudaFree( d_vel);
	cudaFree( d_pos);
	cudaFree( d_final_pos);
	cudaFree( d_rforce);
	cudaFree( d_aforce);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy_per_node);
}
