#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"

//This kernel is extra stupid.  Each block consists of 32 threads which just computes the 
//value of this block.
__global__ void float2_gpu_kernel ( 	unsigned char* graph_data,
					float2* pos, 
					float2* vel,
					float2* final_pos,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
	
		float2 force;
		force.x = 0.0f;
		force.y = 0.0f;
		
		//step 1 - compute force on this node
		for (int i = 0; i < N; ++i)
		{
			if (i != index)
			{				
				float dist_x = pos[i].x - pos[index].x;
				float dist_y = pos[i].y - pos[index].y;
				float dist_squared = dist_x * dist_x + dist_y * dist_y;
				float dist = sqrt(dist_squared);

				force.x += -dist_x * REPULSIVE_FORCE_CONSTANT / dist_squared / dist;
				force.y += -dist_y * REPULSIVE_FORCE_CONSTANT / dist_squared / dist;
				
				force.x += ATTRACTIVE_FORCE_CONSTANT * (dist_x - dist_x * 100 / dist) * graph_data[i * N + index];
				force.y += ATTRACTIVE_FORCE_CONSTANT * (dist_y - dist_y * 100 / dist) * graph_data[i * N + index];
			}
		}	
		vel[index].x += (TIMESTEP * force.x);
		vel[index].y += (TIMESTEP * force.y);
		
		vel[index].x *= DAMPING;
		vel[index].y *= DAMPING;
		
		final_pos[index].x = pos[index].x + TIMESTEP * vel[index].x;
		final_pos[index].y = pos[index].y + TIMESTEP * vel[index].y;			
	}
}

//This kernel is extra stupid.  Each block consists of 32 threads which just computes the 
//value of this block.
static __global__ void compute_energy  ( float2* vel,
					float*  energy,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	if ( index == 0 )
	{
		*energy = 0.0f;
		for (unsigned int i = 0 ; i < N; ++ i )
		{
			*energy += (vel[i].x * vel[i].x + vel[i].y * vel[i].y);
		}
	}
}

extern "C" void float2_fps_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float2 *d_pos, *d_vel, *d_final_pos;
	unsigned char* d_graph_data;
	unsigned char* mapped_graph_data;
	float2 *h_vel, *h_pos;
	float* d_kenetic_energy;
	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_pos, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_vel, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_pos, sizeof(float2) * num_nodes));
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned char) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy, sizeof(float) ) );
	
	//translate all the data into the formats we are using in CUDA.
	mapped_graph_data = (unsigned char*)malloc(num_nodes * num_nodes * sizeof(unsigned char) );
	h_vel = (float2*)malloc(sizeof(float2) * num_nodes);
	h_pos = (float2*)malloc(sizeof(float2) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_vel[i].x = 0.0f;
		h_vel[i].y = 0.0f;
		
		h_pos[i].x = x_pos[i];
		h_pos[i].y = y_pos[i];
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned char) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_vel, (void*) h_vel, sizeof(float2) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_pos, (void*) h_pos, sizeof(float2) * num_nodes, cudaMemcpyHostToDevice));

	int counter = 0;
	float kenetic_energy = 0.0f;
	do 
	{
		float2_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_graph_data, d_pos, d_vel, d_final_pos, num_nodes);						
		cudaThreadSynchronize();

		compute_energy<<< 1, 1 >>>(d_vel, d_kenetic_energy, num_nodes);
		cudaThreadSynchronize();

		CUDA_SAFE_CALL(cudaMemcpy ( (void*) &kenetic_energy, (void*) d_kenetic_energy, sizeof(float), cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		
		//swap the position buffers
		float2 *tmp;
		tmp = d_pos;
		d_pos = d_final_pos;
		d_final_pos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while ( kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);
	
	//restore the points to the correct buffers (d_pos contains the last result because pointers were swapped)
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_pos, (void*) d_pos, sizeof(float2) * num_nodes, cudaMemcpyDeviceToHost));
	for (unsigned int i = 0; i < num_nodes; ++i)
	{
		x_pos[i] = h_pos[i].x;
		y_pos[i] = h_pos[i].y;
	}
	
	
	free(mapped_graph_data);
	free(h_vel);
	cudaFree( d_vel);
	cudaFree( d_pos);
	cudaFree( d_final_pos);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy);
}
