#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"

// This kernel computes the forces on the nodes (each thread computes the forces
// between a pair of nodes)
__global__ void force_calc_kin_gpu_kernel ( 	unsigned char* graph_data,
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N*N)
	{
    	int n1_index = index / N;
	    int n2_index = index - (n1_index * N);
	    
	    if (n1_index == n2_index) {
	       r_force[index].x = 0.0f;
	       r_force[index].y = 0.0f;
	       a_force[index].x = 0.0f;
	       a_force[index].y = 0.0f;
        }
        else {
			// Compute distance between Node1 and Node2
        	float dist_x = pos[n2_index].x - pos[n1_index].x;
			float dist_y = pos[n2_index].y - pos[n1_index].y;
			float dist_squared = dist_x * dist_x + dist_y * dist_y;
			float dist = sqrt(dist_squared);
			
			// Compute Repulsive Force (but without REPULSIVE_FORCE_CONSTANT)
			r_force[index].x = -dist_x / dist_squared / dist;
			r_force[index].y = -dist_y / dist_squared / dist;

			// Compute Attractive Force (but without ATTRACTIVE_FORCE_CONSTANT)
            a_force[index].x = (dist_x - dist_x * 100 / dist) * graph_data[index];
            a_force[index].y = (dist_y - dist_y * 100 / dist) * graph_data[index];
        }
    }
}

// Calculates the position of each node (based on forces already pre-computed)
__global__ void pos_calc_kin_gpu_kernel (
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					float2* vel,
					float2* final_pos,
					float* d_kenetic_energy_per_node,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
		float2 my_r_force;
		float2 my_a_force;
		my_r_force.x = 0.0f;
		my_r_force.y = 0.0f;
		my_a_force.x = 0.0f;
		my_a_force.y = 0.0f;

        // Compute forces on this node
		for (int i = index*N; i < (index+1)*N; i++)
		{
			my_r_force.x += r_force[i].x;
			my_r_force.y += r_force[i].y;
			my_a_force.x += a_force[i].x;
			my_a_force.y += a_force[i].y;
		}
		
		my_r_force.x *= REPULSIVE_FORCE_CONSTANT;
		my_r_force.y *= REPULSIVE_FORCE_CONSTANT;
		my_a_force.x *= ATTRACTIVE_FORCE_CONSTANT;
		my_a_force.y *= ATTRACTIVE_FORCE_CONSTANT;
		
		// Compute Velocities on this node
		vel[index].x += (TIMESTEP * (my_r_force.x + my_a_force.x));
		vel[index].y += (TIMESTEP * (my_r_force.y + my_a_force.y));
		
		vel[index].x *= DAMPING;
		vel[index].y *= DAMPING;
		
		// Compute Final Position of this node
		final_pos[index].x = pos[index].x + TIMESTEP * vel[index].x;
		final_pos[index].y = pos[index].y + TIMESTEP * vel[index].y;

        // Compute the Kinetic Energy of this node        
        d_kenetic_energy_per_node[index] = vel[index].x * vel[index].x + vel[index].y * vel[index].y;
	}
}

extern "C" void force_calc_kin_fps_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float2 *d_rforce, *d_aforce, *d_pos, *d_vel, *d_final_pos;
	unsigned char* d_graph_data;
	unsigned char* mapped_graph_data;
	float2 *h_vel, *h_pos;
	float *d_kenetic_energy, *d_kenetic_energy_per_node;
	float *h_kenetic_energy_per_node;
	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_pos, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_vel, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_pos, sizeof(float2) * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_rforce, sizeof(float2) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_aforce, sizeof(float2) * num_nodes * num_nodes));
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned char) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy_per_node, sizeof(float) * num_nodes) );
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy, sizeof(float) ) );
	h_kenetic_energy_per_node = (float*)malloc(sizeof(float) * num_nodes);
		
	//translate all the data into the formats we are using in CUDA.
	mapped_graph_data = (unsigned char*)malloc(num_nodes * num_nodes * sizeof(unsigned char) );
	h_vel = (float2*)malloc(sizeof(float2) * num_nodes);
	h_pos = (float2*)malloc(sizeof(float2) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_vel[i].x = 0.0f;
		h_vel[i].y = 0.0f;
		
		h_pos[i].x = x_pos[i];
		h_pos[i].y = y_pos[i];
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned char) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_vel, (void*) h_vel, sizeof(float2) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_pos, (void*) h_pos, sizeof(float2) * num_nodes, cudaMemcpyHostToDevice));

	int counter = 0;
	float kenetic_energy = 0.0f;
	do 
	{
		force_calc_kin_gpu_kernel<<< (num_nodes*num_nodes + 31) / 32, 32>>>(d_graph_data, d_rforce, d_aforce, d_pos, num_nodes);						
		cudaThreadSynchronize();

		pos_calc_kin_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_rforce, d_aforce, d_pos, d_vel, d_final_pos, d_kenetic_energy_per_node, num_nodes);						
		cudaThreadSynchronize();

        // Compute the Kinetic Energy of the System
		CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_kenetic_energy_per_node, (void*) d_kenetic_energy_per_node, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		kenetic_energy = 0.0f;
		for (unsigned int i = 0; i < num_nodes; ++i) {
		  kenetic_energy += h_kenetic_energy_per_node[i];
        }
		
		//swap the position buffers
		float2 *tmp;
		tmp = d_pos;
		d_pos = d_final_pos;
		d_final_pos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while ( kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);
	
	//restore the points to the correct buffers (d_pos contains the last result because pointers were swapped)
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_pos, (void*) d_pos, sizeof(float2) * num_nodes, cudaMemcpyDeviceToHost));
	for (unsigned int i = 0; i < num_nodes; ++i)
	{
		x_pos[i] = h_pos[i].x;
		y_pos[i] = h_pos[i].y;
	}
	
	
	free(mapped_graph_data);
	free(h_vel);
	cudaFree( d_vel);
	cudaFree( d_pos);
	cudaFree( d_final_pos);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy);
}
