#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"

// This kernel computes the forces on the nodes (each thread computes the forces
// between a pair of nodes)
__global__ void local_memory_force_gpu_kernel ( 	unsigned char* graph_data,
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;

	if (index < N*N)
	{
		int n1_index = index / N;
		int n2_index = index - (n1_index * N);

		float2 n1_pos = pos[n1_index];
		float2 n2_pos = pos[n2_index];
		unsigned char edge_weight = graph_data[index];

		float2 my_r_force;
		float2 my_a_force;

		if (n1_index == n2_index) {
			my_r_force.x = 0.0f;
			my_r_force.y = 0.0f;
			my_a_force.x = 0.0f;
			my_a_force.y = 0.0f;
		} else {

			// Compute distance between Node1 and Node2
			float dist_x = n2_pos.x - n1_pos.x;
			float dist_y = n2_pos.y - n1_pos.y;
			float dist_squared = dist_x * dist_x + dist_y * dist_y;
			float dist = sqrt(dist_squared);

			// Compute Repulsive Force (but without REPULSIVE_FORCE_CONSTANT)
			my_r_force.x = -dist_x / dist_squared / dist;
			my_r_force.y = -dist_y / dist_squared / dist;

			// Compute Attractive Force (but without ATTRACTIVE_FORCE_CONSTANT)
			my_a_force.x = (dist_x - dist_x * 100 / dist) * edge_weight;
			my_a_force.y = (dist_y - dist_y * 100 / dist) * edge_weight;
		}

		r_force[index] = my_r_force;
		a_force[index] = my_a_force;
	}
}


// Calculates the position of each node (based on forces already pre-computed)
__global__ void local_memory_pos_gpu_kernel (
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					float2* vel,
					float2* final_pos,
					float* d_kenetic_energy_per_node,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
		float2 my_pos = pos[index];
		float2 my_vel = vel[index];
		float2 my_r_force;
		float2 my_a_force;
		my_r_force.x = 0.0f;
		my_r_force.y = 0.0f;
		my_a_force.x = 0.0f;
		my_a_force.y = 0.0f;

		// Compute force on this node
		for (int i = index*N; i < (index+1)*N; ++i)
		{
			float2 r = r_force[i];
			float2 a = a_force[i];
			my_r_force.x += r.x;
			my_r_force.y += r.y;
			my_a_force.x += a.x;
			my_a_force.y += a.y;
		}

		my_r_force.x *= REPULSIVE_FORCE_CONSTANT;
		my_r_force.y *= REPULSIVE_FORCE_CONSTANT;
		my_a_force.x *= ATTRACTIVE_FORCE_CONSTANT;
		my_a_force.y *= ATTRACTIVE_FORCE_CONSTANT;

		// Computer Velocities on this node
		my_vel.x = (my_vel.x + (TIMESTEP * (my_r_force.x + my_a_force.x))) * DAMPING;
		my_vel.y = (my_vel.y + (TIMESTEP * (my_r_force.y + my_a_force.y))) * DAMPING;
		vel[index] = my_vel;

		// Compute Final Position of this node
		float2 my_final_pos;
		my_final_pos.x = my_pos.x + TIMESTEP * my_vel.x;
		my_final_pos.y = my_pos.y + TIMESTEP * my_vel.y;			
		final_pos[index] = my_final_pos;

		// Compute the Kinetic Energy of this node
		d_kenetic_energy_per_node[index] = my_vel.x * my_vel.x + my_vel.y * my_vel.y;
	}
}


extern "C" void local_memory_fps_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float2 *d_rforce, *d_aforce, *d_pos, *d_vel, *d_final_pos;
	unsigned char* d_graph_data;
	unsigned char* mapped_graph_data;
	float2 *h_vel, *h_pos;
	float *d_kenetic_energy_per_node;
	float *h_kenetic_energy_per_node;
	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_pos, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_vel, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_pos, sizeof(float2) * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_rforce, sizeof(float2) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_aforce, sizeof(float2) * num_nodes * num_nodes));
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned char) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy_per_node, sizeof(float) * num_nodes) );
	h_kenetic_energy_per_node = (float*)malloc(sizeof(float) * num_nodes);
		
	//translate all the data into the formats we are using in CUDA.
	mapped_graph_data = (unsigned char*)malloc(num_nodes * num_nodes * sizeof(unsigned char) );
	h_vel = (float2*)malloc(sizeof(float2) * num_nodes);
	h_pos = (float2*)malloc(sizeof(float2) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_vel[i].x = 0.0f;
		h_vel[i].y = 0.0f;
		
		h_pos[i].x = x_pos[i];
		h_pos[i].y = y_pos[i];
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned char) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_vel, (void*) h_vel, sizeof(float2) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_pos, (void*) h_pos, sizeof(float2) * num_nodes, cudaMemcpyHostToDevice));

	int counter = 0;
	float kenetic_energy = 0.0f;
	do 
	{
		local_memory_force_gpu_kernel<<< (num_nodes*num_nodes + 31) / 32, 32>>>(d_graph_data, d_rforce, d_aforce, d_pos, num_nodes);						
		cudaThreadSynchronize();

		local_memory_pos_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_rforce, d_aforce, d_pos, d_vel, d_final_pos, d_kenetic_energy_per_node, num_nodes);
		cudaThreadSynchronize();

		// Compute the Kinetic Energy of the System
		//do_reduction(d_vel, num_nodes, &kenetic_energy);
		CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_kenetic_energy_per_node, (void*) d_kenetic_energy_per_node, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		kenetic_energy = 0.0f;
		for (unsigned int i = 0; i < num_nodes; ++i) {
			kenetic_energy += h_kenetic_energy_per_node[i];
		}

		//swap the position buffers
		float2 *tmp;
		tmp = d_pos;
		d_pos = d_final_pos;
		d_final_pos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while ( kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);
	
	//restore the points to the correct buffers (d_pos contains the last result because pointers were swapped)
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_pos, (void*) d_pos, sizeof(float2) * num_nodes, cudaMemcpyDeviceToHost));
	for (unsigned int i = 0; i < num_nodes; ++i)
	{
		x_pos[i] = h_pos[i].x;
		y_pos[i] = h_pos[i].y;
	}
	
	
	free(h_kenetic_energy_per_node);
	free(mapped_graph_data);
	free(h_vel);
	free(h_pos);
	cudaFree( d_vel);
	cudaFree( d_pos);
	cudaFree( d_final_pos);
	cudaFree( d_rforce);
	cudaFree( d_aforce);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy_per_node);
}
