#include <stdio.h>
#include <cutil.h>
#include <cuda.h>

#include "defines.h"
#include "reduction_force.cu"

// sum up columns instead of rows
#define FORCE_XPOSE 1

// use reduction to add forces, set FORCE_XPOSE to 0 if using reduction
// WARNING! this is extremely slow for some reason...
#define FORCE_REDUCTION 0

// not sure why this slow things down either...
#define FORCE_ALIGN_THREAD_TO_32 0

// This kernel computes the forces on the nodes (each thread computes the forces
// between a pair of nodes)
__global__ void matrix_transpose_force_gpu_kernel ( 	unsigned int* graph_data,
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;

#if FORCE_ALIGN_THREAD_TO_32

	int N_aligned = N*((N+31)/32);
#if FORCE_XPOSE
	int n2_index = index / N_aligned;
	int n1_index = index - (n2_index * N_aligned);
	index = n2_index * N + n1_index;
#else
	int n1_index = index / N_aligned;
	int n2_index = index - (n1_index * N_aligned);
	index = n1_index * N + n2_index;
#endif
	if (n1_index < N && n2_index < N)
	{

#else // !FORCE_ALIGNED_THREAD_TO_32

	if (index < N*N)
	{
#if FORCE_XPOSE
		int n2_index = index / N;
		int n1_index = index - (n2_index * N);
#else
		int n1_index = index / N;
		int n2_index = index - (n1_index * N);
#endif

#endif

		float2 n1_pos = pos[n1_index];
		float2 n2_pos = pos[n2_index];
		unsigned int edge_weight = graph_data[index];

		float2 my_r_force;
		float2 my_a_force;

		if (n1_index == n2_index) {
			my_r_force.x = 0.0f;
			my_r_force.y = 0.0f;
			my_a_force.x = 0.0f;
			my_a_force.y = 0.0f;
		} else {

			// Compute distance between Node1 and Node2
			float dist_x = n2_pos.x - n1_pos.x;
			float dist_y = n2_pos.y - n1_pos.y;
			float dist_squared = dist_x * dist_x + dist_y * dist_y;
			float dist = sqrt(dist_squared);

			// Compute Repulsive Force (but without REPULSIVE_FORCE_CONSTANT)
			my_r_force.x = -dist_x / dist_squared / dist;
			my_r_force.y = -dist_y / dist_squared / dist;

			// Compute Attractive Force (but without ATTRACTIVE_FORCE_CONSTANT)
			my_a_force.x = (dist_x - dist_x * 100 / dist) * edge_weight;
			my_a_force.y = (dist_y - dist_y * 100 / dist) * edge_weight;
		}

		r_force[index] = my_r_force;
		a_force[index] = my_a_force;
	}
}


// Calculates the position of each node (based on forces already pre-computed)
__global__ void matrix_transpose_pos_gpu_kernel (
                    float2* r_force,
                    float2* a_force,
					float2* pos, 
					float2* vel,
					float2* final_pos,
					float* d_kenetic_energy_per_node,
					int N
				   )
{
	int index = (blockIdx.x * blockDim.x) + threadIdx.x;
	
	if (index < N)
	{
		float2 my_pos = pos[index];
		float2 my_vel = vel[index];
#if FORCE_REDUCTION
		float2 my_force = r_force[index];
		my_vel.x = (my_vel.x + TIMESTEP * my_force.x) * DAMPING;
		my_vel.y = (my_vel.y + TIMESTEP * my_force.y) * DAMPING;
#else
		float2 my_r_force;
		float2 my_a_force;
		my_r_force.x = 0.0f;
		my_r_force.y = 0.0f;
		my_a_force.x = 0.0f;
		my_a_force.y = 0.0f;

		// Compute force on this node
#if FORCE_XPOSE
		for (int i = index; i < N*N; i+=N)
#else
		for (int i = index*N; i < (index+1)*N; ++i)
#endif
		{
			float2 r = r_force[i];
			float2 a = a_force[i];
			my_r_force.x += r.x;
			my_r_force.y += r.y;
			my_a_force.x += a.x;
			my_a_force.y += a.y;
		}

		my_r_force.x *= REPULSIVE_FORCE_CONSTANT;
		my_r_force.y *= REPULSIVE_FORCE_CONSTANT;
		my_a_force.x *= ATTRACTIVE_FORCE_CONSTANT;
		my_a_force.y *= ATTRACTIVE_FORCE_CONSTANT;

		// Computer Velocities on this node
		my_vel.x = (my_vel.x + (TIMESTEP * (my_r_force.x + my_a_force.x))) * DAMPING;
		my_vel.y = (my_vel.y + (TIMESTEP * (my_r_force.y + my_a_force.y))) * DAMPING;
#endif
		vel[index] = my_vel;

		// Compute Final Position of this node
		float2 my_final_pos;
		my_final_pos.x = my_pos.x + TIMESTEP * my_vel.x;
		my_final_pos.y = my_pos.y + TIMESTEP * my_vel.y;			
		final_pos[index] = my_final_pos;

		// Compute the Kinetic Energy of this node
		d_kenetic_energy_per_node[index] = my_vel.x * my_vel.x + my_vel.y * my_vel.y;
	}
}


extern "C" void matrix_transpose_fps_gpu_algorithm
(
	unsigned int num_nodes,
	unsigned char** graph,
	float* x_pos,
	float* y_pos
)
{
	float2 *d_rforce, *d_aforce, *d_pos, *d_vel, *d_final_pos;
	unsigned int* d_graph_data;
	unsigned int* mapped_graph_data;
	float2 *h_vel, *h_pos;
	float *d_kenetic_energy_per_node;
	float *h_kenetic_energy_per_node;

	unsigned int p1timer, p2timer;

	cutCreateTimer(&p1timer);
	cutCreateTimer(&p2timer);
	
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_pos, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_vel, sizeof(float2) * num_nodes ));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_final_pos, sizeof(float2) * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_rforce, sizeof(float2) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_aforce, sizeof(float2) * num_nodes * num_nodes));
	
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_graph_data, sizeof(unsigned int) * num_nodes * num_nodes));
	CUDA_SAFE_CALL(cudaMalloc ( (void **)&d_kenetic_energy_per_node, sizeof(float) * num_nodes) );
	h_kenetic_energy_per_node = (float*)malloc(sizeof(float) * num_nodes);
		
	//translate all the data into the formats we are using in CUDA.
	mapped_graph_data = (unsigned int*)malloc(num_nodes * num_nodes * sizeof(unsigned int) );
	h_vel = (float2*)malloc(sizeof(float2) * num_nodes);
	h_pos = (float2*)malloc(sizeof(float2) * num_nodes);
	for (int i = 0; i < num_nodes; ++i)
	{
		for (int j = 0; j < num_nodes; ++j )
		{
			mapped_graph_data[i * num_nodes + j] = graph[i][j];
		}
		h_vel[i].x = 0.0f;
		h_vel[i].y = 0.0f;
		
		h_pos[i].x = x_pos[i];
		h_pos[i].y = y_pos[i];
	}	
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_graph_data, (void*) mapped_graph_data, sizeof(unsigned int) * num_nodes* num_nodes, cudaMemcpyHostToDevice));		
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_vel, (void*) h_vel, sizeof(float2) *num_nodes, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) d_pos, (void*) h_pos, sizeof(float2) * num_nodes, cudaMemcpyHostToDevice));

	int counter = 0;
	float kenetic_energy = 0.0f;

#if FORCE_REDUCTION
	float2 *d_force_per_node;
	CUDA_SAFE_CALL(cudaMalloc ( (void**)&d_force_per_node, sizeof(float2) * num_nodes));
	prepare_reduction_force_memory(num_nodes);
#endif
	do 
	{
#ifdef SHOW_TIME_BREAKDOWN
		cutStartTimer(p1timer);
#endif

#if FORCE_ALIGN_THREAD_TO_32
		int num_nodes_aligned = num_nodes * ((num_nodes + 31)/32);
		matrix_transpose_force_gpu_kernel<<< (num_nodes_aligned * num_nodes_aligned) / 32, 32>>>(d_graph_data, d_rforce, d_aforce, d_pos, num_nodes);
#else
		matrix_transpose_force_gpu_kernel<<< (num_nodes*num_nodes + 31) / 32, 32>>>(d_graph_data, d_rforce, d_aforce, d_pos, num_nodes);						
#endif
		cudaThreadSynchronize();

#if FORCE_REDUCTION
		do_reduction_force(d_rforce, d_aforce, num_nodes, d_force_per_node);
		matrix_transpose_pos_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_force_per_node, d_aforce, d_pos, d_vel, d_final_pos, d_kenetic_energy_per_node, num_nodes);
#else
		matrix_transpose_pos_gpu_kernel<<< (num_nodes + 31) / 32, 32>>>(d_rforce, d_aforce, d_pos, d_vel, d_final_pos, d_kenetic_energy_per_node, num_nodes);
#endif

		cudaThreadSynchronize();

#ifdef SHOW_TIME_BREAKDOWN
	    	cutStopTimer(p1timer);	
		cutStartTimer(p2timer);
#endif
		// Compute the Kinetic Energy of the System
		//do_reduction(d_vel, num_nodes, &kenetic_energy);
		CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_kenetic_energy_per_node, (void*) d_kenetic_energy_per_node, sizeof(float) * num_nodes, cudaMemcpyDeviceToHost));
		cudaThreadSynchronize();
		kenetic_energy = 0.0f;
		for (unsigned int i = 0; i < num_nodes; ++i) {
			kenetic_energy += h_kenetic_energy_per_node[i];
		}
#ifdef SHOW_TIME_BREAKDOWN
	    	cutStopTimer(p2timer);
#endif
		//swap the position buffers
		float2 *tmp;
		tmp = d_pos;
		d_pos = d_final_pos;
		d_final_pos = tmp;
		
		if (DEBUG) if (counter % 10000 == 1) printf("Energy = %f\n", kenetic_energy);
		counter++;
	} while ( kenetic_energy  > STABLE_ENERGY_THRESHOLD * num_nodes);
	
	printf("Total iterations were %d\n", counter);
	printf("Energy = %f\n", kenetic_energy);

#ifdef SHOW_TIME_BREAKDOWN
    	printf("P1 timer time was  %f\n", cutGetTimerValue(p1timer));
    	printf("P2 timer time was  %f\n", cutGetTimerValue(p2timer));
#endif
	
	//restore the points to the correct buffers (d_pos contains the last result because pointers were swapped)
	CUDA_SAFE_CALL(cudaMemcpy ( (void*) h_pos, (void*) d_pos, sizeof(float2) * num_nodes, cudaMemcpyDeviceToHost));
	for (unsigned int i = 0; i < num_nodes; ++i)
	{
		x_pos[i] = h_pos[i].x;
		y_pos[i] = h_pos[i].y;
	}
	
	
	free(h_kenetic_energy_per_node);
	free(mapped_graph_data);
	free(h_vel);
	free(h_pos);
	cudaFree( d_vel);
	cudaFree( d_pos);
	cudaFree( d_final_pos);
	cudaFree( d_rforce);
	cudaFree( d_aforce);
	cudaFree( d_graph_data);
	cudaFree( d_kenetic_energy_per_node);
}
